from __future__ import annotations

import uuid
from datetime import datetime, timezone
from pathlib import Path


def now() -> datetime:
    return datetime.now(timezone.utc)


def create_uuid() -> str:
    return str(uuid.uuid4())


def create_path(path: Path | str) -> Path:
    return path if isinstance(path, Path) else Path(path)


def join_with_dashes(*args) -> str:
    return "-".join(args)


def merge_args(*args, **kwargs) -> dict:
    return {"args": args, "kwargs": kwargs}



from typing import (
    Awaitable,
    Iterable,
    Iterator,
    TypeVar,
)

from tqdm.asyncio import tqdm_asyncio

from agentlens.context import Run

T = TypeVar("T")


def iterate(iterable: Iterable[T], desc: str | None = None) -> Iterator[T]:
    for i, item in enumerate(iterable):
        Run.set_iteration(i)
        yield item
        Run.set_iteration(None)


async def gather(*coros: Awaitable[T], desc: str | None = None) -> list[T]:
    async def eval_coro(i: int, coro: Awaitable[T]) -> T:
        Run.set_iteration(i)
        try:
            return await coro
        finally:
            Run.set_iteration(None)

    tasks = [eval_coro(i, coro) for i, coro in enumerate(coros)]
    return await tqdm_asyncio.gather(*tasks, desc=desc)



from contextlib import contextmanager
from dataclasses import dataclass
from functools import wraps
from logging import getLogger
from pathlib import Path
from typing import (
    Any,
    Callable,
    Coroutine,
    Generator,
    ParamSpec,
    Type,
    TypeVar,
    cast,
    overload,
)

from agentlens.constants import RUNS_DIR
from agentlens.context import Run, Task
from agentlens.evaluation import Hook, HookGenerator, Mock, MockMiss, convert_to_kwargs
from agentlens.serialization import serialize_value, write_json

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R", covariant=True)
F = TypeVar("F", bound=Callable[..., Any])


@contextmanager
def provide(
    *contexts: Any,
    hooks: list[Hook] = [],
    mocks: list[Mock] = [],
) -> Generator[None, None, None]:
    run = Run.current()
    if run is None:
        raise ValueError(
            "No active run context. Use @task decorator or _provide_run() to create one."
        )

    # Create context dictionary
    context_map = {type(context).__name__: context for context in contexts}

    # Check for existing contexts
    current = run.contexts.current
    if current:
        for key in context_map:
            if key in current:
                raise ValueError(f"Context {key} already provided")

    with run.contexts.push(context_map):
        # Create hook dictionary grouped by target
        hook_map: dict[str, list[Hook]] = {}
        for hook in hooks:
            target_name = hook.target.__name__
            if target_name not in hook_map:
                hook_map[target_name] = []
            hook_map[target_name].append(hook)

        # Create mock dictionary grouped by target
        mock_map: dict[str, Mock] = {}
        for mock in mocks:
            if mock.target_name in mock_map:
                raise ValueError(f"Multiple mocks provided for function {mock.target_name}")
            mock_map[mock.target_name] = mock

        with run.hooks.push(hook_map):
            with run.mocks.push(mock_map):
                yield


@contextmanager
def _provide_run(task_name: str) -> Generator[Run, None, None]:
    run = Run.current()
    if run is None:
        run = Run.start(RUNS_DIR, task_name)
        print(f"[agentlens] run directory: {run.dir}")
        yield run
        Run.end()
    else:
        yield run


@overload
def task(fn: F) -> F: ...


@overload
def task(
    *,
    name: str | None = None,
    log_input: bool = True,
    log_output: bool = True,
) -> Callable[[Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]]: ...


def task(
    fn: Callable[P, Coroutine[Any, Any, R]] | None = None,
    *,
    name: str | None = None,
    log_input: bool = True,
    log_output: bool = True,
) -> Callable[[Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]]:
    def decorator(
        fn: Callable[P, Coroutine[Any, Any, R]],
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        if name:
            fn.__name__ = name

        @wraps(fn)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            with _provide_run(fn.__name__) as run:
                with run.create_observation(fn.__name__) as observation:
                    # Convert args to kwargs
                    all_kwargs = convert_to_kwargs(fn, args, kwargs)

                    # Save input if requested
                    if log_input:
                        input_data = {k: serialize_value(v) for k, v in all_kwargs.items()}
                        write_json(input_data, observation.dir / "input.json")

                    # Get hooks for this function
                    hooks = run.hooks.current.get(fn.__name__, []) if run.hooks.current else []

                    # run pre-hooks
                    generators: list[HookGenerator] = []
                    injected_kwargs: dict[str, Any] = {}
                    for hook in hooks:
                        gen = hook(args, kwargs)
                        if isinstance(gen, Generator):
                            generators.append(gen)  # type: ignore[arg-type]
                            new_injected_kwargs = next(gen)
                            injected_kwargs.update(new_injected_kwargs)

                    # rewrite task args/kwargs
                    all_kwargs = convert_to_kwargs(fn, args, kwargs)
                    all_kwargs.update(injected_kwargs)

                    # execute task
                    mock_fn = run.mocks.current.get(fn.__name__) if run.mocks.current else None
                    try:
                        if mock_fn is not None:
                            result = await mock_fn(**all_kwargs)
                        else:
                            raise MockMiss
                    except MockMiss:
                        result = await fn(**all_kwargs)

                    # Save output if requested
                    if log_output:
                        output_data = {"result": serialize_value(result)}
                        write_json(output_data, observation.dir / "output.json")

                    cost = observation.cost
                    if cost.input > 0 or cost.output > 0:
                        (observation.dir / "cost.txt").write_text(str(cost))

                    # send result to generator hooks
                    for gen in generators:
                        try:
                            gen.send(result)
                        except StopIteration:
                            pass

                    return result

        return wrapper

    if fn is not None:
        return decorator(fn)  # type: ignore[return-value]
    else:
        return decorator


@dataclass
class Lens:
    _log = getLogger("agentlens")
    _runs_dir: Path = RUNS_DIR

    def __bool__(self) -> bool:
        return Run.current() is not None

    @property
    def run(self) -> Run:
        run = Run.current()
        if run is None:
            raise ValueError("No run context found")
        return run

    @property
    def task(self) -> Task:
        return Run.current_task()

    def __getitem__(self, context_type: Type[T]) -> T:
        key = context_type.__name__
        contexts = self.run.contexts.current
        if contexts is None:
            raise ValueError(
                f"No context value provided for type {context_type.__name__}. "
                "Use 'with lens.provide(value):' to provide one."
            )
        try:
            return cast(T, contexts[key])
        except KeyError:
            raise ValueError(
                f"No context value provided for type {context_type.__name__}. "
                "Use 'with lens.provide(value):' to provide one."
            )

    def __setitem__(self, context_type: Type[T], value: T) -> None:
        """Set a new value for an existing context."""
        key = context_type.__name__
        contexts = self.run.contexts.current
        if contexts is None:
            raise ValueError(
                f"No context value provided for type {context_type.__name__}. "
                "Use 'with lens.provide(value):' to provide one."
            )
        if key not in contexts:
            raise ValueError(
                f"No context value provided for type {context_type.__name__}. "
                "Use 'with lens.provide(value):' to provide one."
            )
        contexts[key] = value

    def increase_cost(self, *, input: float = 0.0, output: float = 0.0) -> None:
        """Increase cost for current observation and all parent observations"""
        run = self.run
        # Get all observations in the stack, from current to root
        for observation in run.observations.stack:
            observation.cost.input += input
            observation.cost.output += output


lens = Lens()



from typing import Any, Type, TypeVar, cast

import anthropic
from anthropic.resources.messages import Message as AnthropicMessage
from pydantic import BaseModel

from agentlens import lens
from agentlens.message import TextContent
from agentlens.model import ModelUsage
from agentlens.provider import Message, Provider

T = TypeVar("T", bound=BaseModel)


class Anthropic(Provider):
    # Cost per million tokens in USD
    MODEL_COSTS = {
        "claude-3-5-sonnet-20240620": (3.00, 15.00),
        "claude-3-5-sonnet-20241022": (3.00, 15.00),
        "claude-3-opus-20240229": (15.00, 75.00),
        "claude-3-sonnet-20240229": (3.00, 15.00),
        "claude-3-haiku-20240307": (0.25, 1.25),
    }

    def __init__(
        self,
        api_key: str | None = None,
        max_connections: dict[str, int] | None = None,
        max_connections_default: int = 10,
    ):
        super().__init__(
            name="anthropic",
            max_connections=max_connections,
            max_connections_default=max_connections_default,
        )
        self.client = anthropic.AsyncAnthropic(api_key=api_key)

    def get_token_costs(self, model: str) -> tuple[float, float]:
        """Returns (input_cost_per_million, output_cost_per_million)"""
        return self.MODEL_COSTS.get(model, (0.0, 0.0))

    def _extract_usage(self, response: Any) -> ModelUsage:
        usage = getattr(response, "usage", None)
        if usage is None:
            raise ValueError("No usage data found in Anthropic response")
        return ModelUsage(
            input_tokens=usage.input_tokens,
            output_tokens=usage.output_tokens,
            total_tokens=usage.input_tokens + usage.output_tokens,
        )

    def _update_cost(
        self,
        model: str,
        response: Any,
    ) -> None:
        usage = self._extract_usage(response)
        input_cost_per_token, output_cost_per_token = self.get_token_costs(model)
        lens.increase_cost(
            input=usage.input_tokens * input_cost_per_token / 1_000_000,
            output=usage.output_tokens * output_cost_per_token / 1_000_000,
        )

    async def generate_text(
        self,
        *,
        model: str,
        messages: list[Message],
        max_tokens: int = 8096,
        **kwargs,
    ) -> str:
        # Extract and combine system messages
        system_messages = [m for m in messages if m.role == "system"]
        other_messages = [m for m in messages if m.role != "system"]

        # Combine system message texts
        system_text = ""
        for msg in system_messages:
            if isinstance(msg.content, str):
                system_text += msg.content + "\n"
            else:
                for content in msg.content:
                    if isinstance(content, TextContent):
                        system_text += content.text + "\n"

        # Add system parameter to kwargs if we have system messages
        if system_text:
            kwargs["system"] = system_text.strip()

        untyped_message = await self.client.messages.create(
            model=model,
            messages=[m.model_dump() for m in other_messages],  # type: ignore
            max_tokens=max_tokens,
            **kwargs,
        )

        message = cast(AnthropicMessage, untyped_message)

        self._update_cost(model, message)
        assert message.content is not None
        result = message.content[0]
        assert result.type == "text"
        return result.text

    async def generate_object(
        self,
        *,
        model: str,
        messages: list[Message],
        schema: Type[T],
        **kwargs,
    ) -> T:
        raise NotImplementedError("Anthropic does not support object generation")



from typing import Any, Type, TypeVar

from openai import AsyncOpenAI
from pydantic import BaseModel

from agentlens.client import lens
from agentlens.model import ModelUsage
from agentlens.provider import Message, Provider

T = TypeVar("T", bound=BaseModel)


class OpenAI(Provider):
    # Cost per million tokens in USD
    MODEL_COSTS = {
        "gpt-4o-mini": (0.150, 0.600),
        "gpt-4o-mini-2024-07-18": (0.150, 0.600),
        "gpt-4o": (5.00, 15.00),
        "gpt-4o-2024-08-06": (2.50, 10.00),
        "gpt-4o-2024-05-13": (5.00, 15.00),
        "o1-mini": (3.00, 12.00),
        "o1-mini-2024-09-12": (3.00, 12.00),
        "o1-preview": (15.00, 60.00),
        "o1-preview-2024-09-12": (15.00, 60.00),
    }

    def __init__(
        self,
        api_key: str | None = None,
        max_connections: dict[str, int] | None = None,
        max_connections_default: int = 10,
    ):
        super().__init__(
            name="openai",
            max_connections=max_connections,
            max_connections_default=max_connections_default,
        )
        self.client = AsyncOpenAI(api_key=api_key)

    def get_token_costs(self, model: str) -> tuple[float, float]:
        """Returns (input_cost_per_million, output_cost_per_million)"""
        return self.MODEL_COSTS.get(model, (0.0, 0.0))

    def _extract_usage(self, response: Any) -> ModelUsage:
        usage = getattr(response, "usage", None)
        if usage is None:
            raise ValueError("No usage data found in OpenAI response")
        return ModelUsage(
            input_tokens=usage.prompt_tokens,
            output_tokens=usage.completion_tokens,
            total_tokens=usage.total_tokens,
        )

    def _update_cost(
        self,
        model: str,
        response: Any,
    ) -> None:
        usage = self._extract_usage(response)
        input_cost_per_token, output_cost_per_token = self.get_token_costs(model)
        lens.increase_cost(
            input=usage.input_tokens * input_cost_per_token / 1_000_000,
            output=usage.output_tokens * output_cost_per_token / 1_000_000,
        )

    async def generate_text(
        self,
        *,
        model: str,
        messages: list[Message],
        **kwargs,
    ) -> str:
        completion = await self.client.chat.completions.create(
            model=model,
            messages=[m.model_dump() for m in messages],  # type: ignore
            **kwargs,
        )

        self._update_cost(model, completion)
        assert completion.choices[0].message.content is not None
        return completion.choices[0].message.content

    async def generate_object(
        self,
        *,
        model: str,
        messages: list[Message],
        schema: Type[T],
        **kwargs,
    ) -> T:
        completion = await self.client.beta.chat.completions.parse(
            model=model,
            messages=[message.model_dump() for message in messages],  # type: ignore
            response_format=schema,
            **kwargs,
        )

        self._update_cost(model, completion)
        assert completion.choices[0].message.parsed is not None
        return completion.choices[0].message.parsed



from .client import lens, provide, task
from .dataset import Dataset
from .evaluation import HookGenerator
from .inference import generate_object, generate_text
from .iterate import gather, iterate
from .message import (
    Message,
    assistant_message,
    image_content,
    system_message,
    user_message,
)
from .provider import Model, Provider
from .provider_anthropic import Anthropic
from .provider_openai import OpenAI

__all__ = [
    "AI",
    "OpenAI",
    "Anthropic",
    "Provider",
    "Model",
    "Dataset",
    "lens",
    "client",
    "HookGenerator",
    "task",
    "provide",
    "Message",
    "system_message",
    "user_message",
    "assistant_message",
    "image_content",
    "generate_object",
    "generate_text",
    "gather",
    "iterate",
]
